#include <assert.h>
#include <stdlib.h>

#include "jebelean.h"

#include "trace.h"

void
trace_array(const char *label, const struct jhgcd *hgcd);

#define SIZE_IN_BITS(x) (mpz_sgn((x)) ? mpz_sizeinbase((x), 2) : 0)

/* m is the size threshold for the remainders. l is the size limit for
   the matrix entries.

   We must have m >= n/2, in order to use Jebelean's criterion. Using
   a slightly larger m than necessary gives us stricter lower bounds
   when the "large" remainders are reconstructed after a recursive
   hgcd call.
*/

#if 1
#define CHOOSE_M(n) (((n) + 1)/2)
#define CHOOSE_L(n) ((n)/2)
#else
#define CHOOSE_M(n) ((n)/2+1)
#define CHOOSE_L(n) (((n)+1)/2-1)
#endif

/* Quotient stack */
void
quotient_stack_init(struct quotient_stack *stack)
{
  stack->top = NULL;
  stack->bottom = NULL;
}

void
quotient_stack_clear(struct quotient_stack *stack)
{
  struct quotient_node *p;
  struct quotient_node *old;

  for (old = stack->top; old; old = p)
    {
      mpz_clear(old->q);
      p = old->next;
      free(old);
    }

  stack->top = NULL;
  stack->bottom = NULL;
}

void
quotient_stack_push(struct quotient_stack *stack,
		    const mpz_t q)
{
  struct quotient_node *n = malloc(sizeof(*n));
  if (!n)
    abort();

  mpz_init_set(n->q, q);
  n->next = stack->top;    
  stack->top = n;
  if (!stack->bottom)
    stack->bottom = n;
}

void
quotient_stack_pop(struct quotient_stack *stack,
		   mpz_t q)
{
  struct quotient_node *n = stack->top;
  if (!n)
    abort();

  stack->top = n->next;
  mpz_swap(n->q, q);
  mpz_clear(n->q);
  free(n);

  if (!stack->top)
    stack->bottom = NULL;
}

void
quotient_stack_nappend(struct quotient_stack *R,
		       struct quotient_stack *A, struct quotient_stack *B)
{
  assert(R != A);
  assert(R != B);
  assert(B != A);

  assert(!R->top);
  assert(!R->bottom);

  if (A->top)
    {
      R->top = A->top; R->bottom = A->bottom;
      A->top = NULL; A->bottom = NULL;

      if (B->top)
	{
	  R->bottom->next = B->top; R->bottom = B->bottom;
	  B->top = NULL; B->bottom = NULL;
	}
    }
  else
    {
      R->top = B->top; R->bottom = B->bottom;
      B->top = NULL; B->bottom = NULL;
    }
}

void
jhgcd_init(struct jhgcd *hgcd)
{
  unsigned i;
  
  mpz_init(hgcd->a);
  mpz_init(hgcd->b);
  quotient_stack_init(&hgcd->quotients);
  
  for (i = 0; i<4; i++)
    mpz_init(hgcd->R[i]);

  hgcd->sign = 0;
}

void
jhgcd_clear(struct jhgcd *hgcd)
{
  unsigned i;
  
  mpz_clear(hgcd->a);
  mpz_clear(hgcd->b);
  quotient_stack_clear(&hgcd->quotients);
  
  for (i = 0; i<4; i++)
    mpz_clear(hgcd->R[i]);
}

void
jhgcd_check(struct jhgcd *hgcd)
{
  unsigned i;
  mpz_t t;

  mpz_init(t);
  
  assert(mpz_sgn(hgcd->R[0]) >= 0);
  assert(mpz_sgn(hgcd->R[1]) >= 0);
  assert(mpz_sgn(hgcd->R[2]) >= 0);
  assert(mpz_sgn(hgcd->R[3]) >= 0);

  for (i = 0; i<4; i++)
    assert(mpz_sgn(hgcd->R[i]) >= 0);

  if (hgcd->sign >= 0)
    {
      assert(mpz_cmp(hgcd->b, hgcd->R[2]) >= 0);
      /* Check r - r' + v - v' >= 0 */
      mpz_set(t, hgcd->a);
      mpz_sub(t, t, hgcd->b);
      mpz_add(t, t, hgcd->R[3]);
      mpz_sub(t, t, hgcd->R[1]);
      assert(mpz_sgn(t) >= 0);
    }
  else
    {
      assert(mpz_cmp(hgcd->b, hgcd->R[3]) >= 0);
      /* Check r - r' + u - u' >= 0 */
      mpz_set(t, hgcd->a);
      mpz_sub(t, t, hgcd->b);
      mpz_add(t, t, hgcd->R[2]);
      mpz_sub(t, t, hgcd->R[0]);
      assert(mpz_sgn(t) >= 0);
    }
  mpz_clear(t);
}

void
jhgcd_check_more(struct jhgcd *hgcd,
		 const mpz_t a, const mpz_t b)
{
  mpz_t x;
  mpz_t y;

  unsigned n = SIZE_IN_BITS(a);
  unsigned l = CHOOSE_L(n);
  unsigned i;
  
  mpz_init(x);
  mpz_init(y);

  jhgcd_check(hgcd);

  for (i = 0; i<4; i++)
    assert(SIZE_IN_BITS(hgcd->R[i]) <= l);
  
  mpz_mul(x, hgcd->R[0], a);
  mpz_mul(y, hgcd->R[1], b);

  if (hgcd->sign >= 0)
    mpz_sub(x, x, y);
  else
    mpz_sub(x, y, x);

  assert(mpz_cmp(x, hgcd->a) == 0);

  mpz_mul(x, hgcd->R[2], a);
  mpz_mul(y, hgcd->R[3], b);

  if (hgcd->sign >= 0)
    mpz_sub(x, y, x);
  else
    mpz_sub(x, x, y);

  assert(mpz_cmp(x, hgcd->b) == 0);
  
  mpz_clear(x);
  mpz_clear(y);
}

static void
jhgcd_update(struct jhgcd *hgcd, const mpz_t q)
{
  unsigned i;

  mpz_t t;
  mpz_init(t);

  for (i = 0; i<2; i++)
    {
      /* ( |u_{k+2}|, |v_{k+2}| )
	   = ( |u_k|, |v_k| ) + q ( |u_{k+1}|, |v_{k+1}| )
      */
      mpz_mul(t, q, hgcd->R[i+2]);
      mpz_add(hgcd->R[i], hgcd->R[i], t);
      mpz_swap(hgcd->R[i], hgcd->R[i+2]);
    }
  mpz_clear(t);
  hgcd->sign = ~hgcd->sign;

  quotient_stack_push(&hgcd->quotients, q);
}

/* Computes R = A * B. Modifies A and B.  */
static void
jhgcd_mult(struct jhgcd *R, struct jhgcd *A, struct jhgcd *B)
{
  unsigned r;
  unsigned c;
  mpz_t t;
  
  assert(R != A);
  assert(R != B);
  assert(A != B);

  mpz_init(t);
#if 0
  trace_array("jhgcd_mult: A", A);
  trace_array("jhgcd_mult: B", B);
#endif
  /* One element of R at a time */
  for (r = 0; r<2; r++)
    for (c = 0; c<2; c++)
      {
	mpz_mul(R->R[2*r+c], A->R[2*r], B->R[c]);
	mpz_mul(t, A->R[2*r+1], B->R[2+c]);
	mpz_add(R->R[2*r+c], R->R[2*r+c], t);
      }

  R->sign = A->sign ^ B->sign;
  quotient_stack_nappend(&R->quotients, &A->quotients, &B->quotients);
#if 0
  trace_array("jhgcd_mult: R", R);
#endif
  mpz_clear(t);
}
     
static int
jhgcd_condition(const struct jhgcd *hgcd,
		const mpz_t u, const mpz_t v,
		const mpz_t r)
{
  mpz_t t;
  mpz_init(t);
  
  mpz_sub(t, hgcd->b, r);
  if (hgcd->sign < 0)
    {
      /* Check if a_{k+1} - a{k+2} >= v_{k+2} - v{k+1} (k+1 even)
	 where

	 a_{k+1} = hgcd->b,
	 a_{k+2} = r,
	 v_{k+1} = -hgcd->R[3],
	 v_{k+2} = v_k - q[0] v_{k+1}
	         = hgcd->R[1] + q[0] * hgcd->R[3])
		 = v
      */

      mpz_sub(t, t, v);
      mpz_sub(t, t, hgcd->R[3]);
      if (mpz_sgn(t) < 0)
	{
	no:
	  mpz_clear(t);
	  return 0;
	}
    }
  else
    {
      /* Check if a_{k+1} - a{k+2} >= u_{k+2} - u{k+1} (k+1 odd)
	 where
	 
	 a_{k+1} = hgcd->b,
	 a_{k+2} = r,
	 u_{k+1} = -hgcd->R[2],
	 u_{k+2} = u_k - q[0] u_{k+1}
	         = hgcd->R[0] + q[0] * hgcd->R[2])
		 = u
      */
      mpz_sub(t, t, u);
      mpz_sub(t, t, hgcd->R[2]);
      if (mpz_sgn(t) < 0)
	goto no;
    }
  mpz_clear(t);
  return 1;
}

/* Returns an array (u,v;u',v') with elements of size at most
 * l = floor(#a/2), satisfying Jebelean's criterion:
 *
 *   r = u a + v b
 *   r'= u'a + v'b
 *
 *   r' >= -u', r - r' >= v'-v  (for sign >= 0)
 *   r' >= -v', r - r' >= u'-u  (for sign < 0)
 *
 * Then the quotient and remainder sequences match up to r, and r'.
 * Also, r must be of size at least m = ceil(n/2).
 */

int
jhgcd_slow(struct jhgcd *hgcd)
{
  unsigned n = SIZE_IN_BITS(hgcd->a);

  /* m >= n/2 is needed in order to guarantee that |v'| <= a_{i+1} */
  unsigned m = CHOOSE_M(n);
  unsigned l = CHOOSE_L(n);
  
  unsigned i;

  trace("jhgcd_slow: n = %d, m = %d, l = %d\n",
	n, m, l);
  
  /* Maintain r[0], q[0] quotient and remainder of a and b, r[1], q[1]
   * quotient and remainder of b and r[0].
   *
   * We should have a, b, r[0] >= 2^m, and loop until r[1] < 2^m. */
  mpz_t r[2];
  mpz_t q[2];
  mpz_t u;
  mpz_t v;

  mpz_set_ui(hgcd->R[0], 1);
  mpz_set_ui(hgcd->R[1], 0);
  mpz_set_ui(hgcd->R[2], 0);
  mpz_set_ui(hgcd->R[3], 1);
  
  if (SIZE_IN_BITS(hgcd->b) < m)
    return 0;

  mpz_init(r[0]);
  mpz_init(q[0]);
  
  mpz_fdiv_qr(q[0], r[0], hgcd->a, hgcd->b);
  if (SIZE_IN_BITS(r[0]) < m)
    {
      mpz_clear(r[0]);
      mpz_clear(q[0]);
      return 0;
    }
  
  mpz_init(r[1]);
  mpz_init(q[1]);
    
  for (;;)
    {
#if 0
      trace("jhgcd_slow: R = %s(%Zd,%Zd;%Zd,%Zd)\n"
	    "  a = %Zd, b = %Zd,\n"
	    "  r[0] = %Zd, q[0] = %Zd\n",
	    hgcd->sign < 0 ? "-" : "",
	    hgcd->R[0], hgcd->R[1], hgcd->R[2], hgcd->R[3],
	    hgcd->a, hgcd->b,
	    r[0], q[0]);
#endif
      mpz_fdiv_qr(q[1], r[1], hgcd->b, r[0]);
#if 0
      trace("  r[1] = %Zd, q[1] = %Zd\n",
	    r[1], q[1]);
#endif
      if (SIZE_IN_BITS(r[1]) < m)
	break;

      jhgcd_update(hgcd, q[0]);
      mpz_swap(hgcd->a, r[1]);
      mpz_swap(hgcd->a, hgcd->b);
      mpz_swap(hgcd->b, r[0]);

      mpz_swap(q[0], q[1]);
    }

  /* Now a,b,r[0] >= 2^m, r[1] < 2^m */
  assert(SIZE_IN_BITS(hgcd->a) >= m);
  assert(SIZE_IN_BITS(hgcd->b) >= m);
  assert(SIZE_IN_BITS(r[0]) >= m);
  assert(SIZE_IN_BITS(r[1]) < m);

  for(i = 0; i<4; i++)
    /* As a_0 < 2^n and hgcd->a >= 2^m, we should have
       
       |u|, |v| <= a_0 / a < 2^{n - m}
                 = 2^floor(n/2)
    */
    assert(SIZE_IN_BITS(hgcd->R[i]) <= l);
  
  mpz_init(u);
  mpz_init(v);

  mpz_mul(u, q[0], hgcd->R[2]);
  mpz_add(u, u, hgcd->R[0]);

  mpz_mul(v, q[0], hgcd->R[3]);
  mpz_add(v, v, hgcd->R[1]);
  
  if (SIZE_IN_BITS(u) > l
      || SIZE_IN_BITS(v) > l
      || !jhgcd_condition(hgcd, u, v, r[0]))
    goto done;
  
  /* Ok, we can use r[0]. */
  trace("jhgcd_slow: r[0] can be used.\n");
  
  jhgcd_update(hgcd, q[0]); /* Recomputes u' and v' */
  assert(mpz_cmp(u, hgcd->R[2]) == 0);
  assert(mpz_cmp(v, hgcd->R[3]) == 0);

  mpz_swap(hgcd->a, r[0]);
  mpz_swap(hgcd->a, hgcd->b);

  /* Check r[1] */
  mpz_mul(u, q[1], hgcd->R[2]);
  mpz_add(u, u, hgcd->R[0]);

  mpz_mul(v, q[1], hgcd->R[3]);
  mpz_add(v, v, hgcd->R[1]);

  if (hgcd->sign >= 0)
    {
      /* r[0], u >= 0, r[1], v  <= 0 */
      if (mpz_cmp(r[1], v) < 0)
	goto done;
    }
  else
    {
      /* r[0], u <= 0, r[1], v >= 0 */
      if (mpz_cmp(r[1], u) < 0)
	goto done;
    }

  if (SIZE_IN_BITS(u) > l
      || SIZE_IN_BITS(v) > l
      || !jhgcd_condition(hgcd, u, v, r[1]))
    goto done;

  /* Ok, we can use r[1] too. */
  trace("jhgcd_slow: r[1] can be used.\n");
  jhgcd_update(hgcd, q[1]); /* Recomputes u' and v' */
  assert(mpz_cmp(u, hgcd->R[2]) == 0);
  assert(mpz_cmp(v, hgcd->R[3]) == 0); 

  mpz_swap(hgcd->a, r[1]);
  mpz_swap(hgcd->a, hgcd->b);
  
 done:
  mpz_clear(r[0]);
  mpz_clear(r[1]);
  mpz_clear(q[0]);
  mpz_clear(q[1]);
  mpz_clear(u);
  mpz_clear(v);

  return 1;
}

static void
jhgcd_vector(struct jhgcd *hgcd,
	     mpz_t a, mpz_t b)
{
  mpz_t s;
  mpz_t t;
  mpz_init(s);
  mpz_init(t);

  mpz_mul(s, a, hgcd->R[0]);
  mpz_mul(t, b, hgcd->R[1]);
  if (hgcd->sign < 0)
    mpz_sub(t, t, s);
  else
    mpz_sub(t, s, t);

  assert(mpz_sgn(t) >= 0);

  mpz_mul(s, a, hgcd->R[2]);
  mpz_swap(t, a);
  mpz_mul(t, b, hgcd->R[3]);

  if (hgcd->sign < 0)
    mpz_sub(b, s, t);
  else
    mpz_sub(b, t, s);

  mpz_clear(s);
  mpz_clear(t);
}

static void
jhgcd_backup(struct jhgcd *hgcd)
{
  unsigned i;
  mpz_t q; 
  mpz_t t;
  mpz_init(q);
  mpz_init(t);
  
  quotient_stack_pop(&hgcd->quotients, q);
  trace("jhgcd_backup: q = %Zd\n", q);
  
  /* Multiply R by

     (0, 1; 1, -q)^-1 = (q, 1; 1, 0) == (0, 1; 1, 0) (1, 0; q, 1)

     i.e add q a to b, then swap. */

  /* Example: a = 8, b = 5, q = 1
     b <- qa + b = 13, a = 8 */
     
  mpz_mul(t, hgcd->a, q);
  mpz_add(hgcd->b, hgcd->b, t);
  mpz_swap(hgcd->a, hgcd->b);

  /* Same for the R matrix, one column at a time */
  for (i = 0; i<2; i++)
    {
      mpz_mul(t, hgcd->R[i], q);
      mpz_sub(hgcd->R[i+2], hgcd->R[i+2], t);
      assert(mpz_sgn(hgcd->R[i+2]) >= 0);
      
      mpz_swap(hgcd->R[i], hgcd->R[i+2]);
    }

  hgcd->sign = ~hgcd->sign;
  trace_array("jhgcd_backup", hgcd);
}

static int
jhgcd_real(struct jhgcd *hgcd);

int
jhgcd(struct jhgcd *hgcd)
{
  mpz_t a;
  mpz_t b;
  int res;
  
  mpz_init_set(a, hgcd->a);
  mpz_init_set(b, hgcd->b);

  res = jhgcd_real(hgcd);
  jhgcd_check_more(hgcd, a, b);
  mpz_clear(a);
  mpz_clear(b);

  return res;
}

int
jhgcd_real(struct jhgcd *hgcd)
{
  unsigned n = SIZE_IN_BITS(hgcd->a);
  unsigned m = CHOOSE_M(n);
  unsigned l = CHOOSE_L(n);
  unsigned k;
  unsigned i;
  
  struct jhgcd R;
  struct jhgcd S;
  mpz_t q[2];
  mpz_t r[2];	  
  mpz_t u;
  mpz_t v;

  static int entry = 0;
  int count = ++entry;

#if 0
  trace("jhgcd(%d): a = %Zd, b = %Zd\n"
	"  n = %d, m = %d, l = %d\n",
	count, hgcd->a, hgcd->b, n, m, l);
#endif
  if (n < 5)
    return jhgcd_slow(hgcd);

  mpz_set_ui(hgcd->R[0], 1);
  mpz_set_ui(hgcd->R[1], 0);
  mpz_set_ui(hgcd->R[2], 0);
  mpz_set_ui(hgcd->R[3], 1);
  
  if (SIZE_IN_BITS(hgcd->b) < m)
    return 0;

  k = m;
  assert(k);

  jhgcd_init(&R);
  jhgcd_init(&S);
  
  mpz_fdiv_q_2exp(R.a, hgcd->a, k);
  mpz_fdiv_q_2exp(R.b, hgcd->b, k);

  if (!jhgcd(&R))
    {
      trace("jhgcd(%d): recursive jhgcd call returned 0. n = %d\n"
	    "  Falling back on jhgcd_slow.\n",
	    count, n);
      return jhgcd_slow(hgcd);
    }

  jhgcd_vector(&R, hgcd->a, hgcd->b);
  
  assert(mpz_sgn(hgcd->a) >= 0);
  assert(mpz_sgn(hgcd->b) >= 0);
  
  assert(SIZE_IN_BITS(hgcd->a) >= m);
  assert(SIZE_IN_BITS(hgcd->b) >= m);

  assert(mpz_cmp(hgcd->a, hgcd->b) > 0);
#if 0
  if (mpz_cmp(hgcd->a, hgcd->b) < 0
      || SIZE_IN_BITS(hgcd->b) < m)
    jhgcd_backup(hgcd);

  assert(SIZE_IN_BITS(hgcd->a) >= m + k/2);
  assert(SIZE_IN_BITS(hgcd->b) >= m);
#endif
  
  n = SIZE_IN_BITS(hgcd->a);
#if 0
  k = 2*m - n;
#else
  /* FIXME: This seems to be the way to guarantee that #a >= m. An
     alternative way is to require that m >= l + 1, not just m >= l
     like we do now (m = (n+1)/2, l = n-m). */
     
  k = m;
#endif
  mpz_fdiv_q_2exp(S.a, hgcd->a, k);
  mpz_fdiv_q_2exp(S.b, hgcd->b, k);
  
  if (jhgcd(&S))
    {
      jhgcd_vector(&S, hgcd->a, hgcd->b);
      jhgcd_mult(hgcd, &S, &R);
      
      assert(mpz_sgn(hgcd->a) >= 0);
      assert(mpz_sgn(hgcd->b) >= 0);
      
      assert(SIZE_IN_BITS(hgcd->a) >= m);
      assert(SIZE_IN_BITS(hgcd->b) >= m);
      assert(mpz_cmp(hgcd->a, hgcd->b) >= 0);
#if 0
      if (mpz_cmp(hgcd->a, hgcd->b) < 0
	  || SIZE_IN_BITS(hgcd->b) < m)
	/* It might be a little more efficient to do the backup
	   operation on S rather than on hgcd, but that makes the
	   book-keeping more compicated. */
	jhgcd_backup(hgcd);
#endif 
      assert(SIZE_IN_BITS(hgcd->a) >= m);
      assert(SIZE_IN_BITS(hgcd->b) >= m);	  
    }
  else
    {
      trace("jhgcd(%d): Second recursive called returned 0\n", count);
      /* Relies on S being initialized to the identity */
      jhgcd_mult(hgcd, &S, &R);
    }
  
  mpz_init(q[0]);
  mpz_init(q[1]);
  mpz_init(r[0]);
  mpz_init(r[1]);
  
  mpz_fdiv_qr(q[0], r[0], hgcd->a, hgcd->b);
  if (SIZE_IN_BITS(r[0]) < m)
    {
      trace("jhgcd(%d): r[0] is too small, backing up.\n", count);
      mpz_swap(r[0], r[1]);
      mpz_swap(q[0], q[1]);
      
      mpz_set(q[0], hgcd->quotients.top->q);
      mpz_set(r[0], hgcd->b);  
      jhgcd_backup(hgcd);
    }
  else for (;;)
    {
#if 0
      trace("jhgcd(%d): R = %s(%Zd,%Zd;%Zd,%Zd)\n"
	    "  a = %Zd, b = %Zd,\n"
	    "  r[0] = %Zd, q[0] = %Zd\n",
	    count, hgcd->sign < 0 ? "-" : "",
	    hgcd->R[0], hgcd->R[1], hgcd->R[2], hgcd->R[3],
	    hgcd->a, hgcd->b,
	    r[0], q[0]);
#endif 
      mpz_fdiv_qr(q[1], r[1], hgcd->b, r[0]);
#if 0
      trace("  r[1] = %Zd, q[1] = %Zd\n",
	    r[1], q[1]);
#endif
      if (SIZE_IN_BITS(r[1]) < m)
	break;
      
      jhgcd_update(hgcd, q[0]);
      mpz_swap(hgcd->a, r[1]);
      mpz_swap(hgcd->a, hgcd->b);
      mpz_swap(hgcd->b, r[0]);

      mpz_swap(q[0], q[1]);
    }

  /* Now a,b,r[0] >= 2^m, r[1] < 2^m */
  assert(SIZE_IN_BITS(hgcd->a) >= m);
  assert(SIZE_IN_BITS(hgcd->b) >= m);
  assert(SIZE_IN_BITS(r[0]) >= m);
  assert(SIZE_IN_BITS(r[1]) < m);

  for(i = 0; i<4; i++)
    /* As a_0 < 2^n and hgcd->a >= 2^m, we should have
       
       |u|, |v| <= a_0 / a < 2^{n - m}
                 = 2^floor(n/2)
    */
    assert(SIZE_IN_BITS(hgcd->R[i]) <= l);
  
  mpz_init(u);
  mpz_init(v);

  mpz_mul(u, q[0], hgcd->R[2]);
  mpz_add(u, u, hgcd->R[0]);

  mpz_mul(v, q[0], hgcd->R[3]);
  mpz_add(v, v, hgcd->R[1]);
  
  if (SIZE_IN_BITS(u) > l
      || SIZE_IN_BITS(v) > l
      || !jhgcd_condition(hgcd, u, v, r[0]))
    goto done;
  
  /* Ok, we can use r[0]. */
  trace("jhgcd(%d): r[0] can be used.\n", count);
  
  jhgcd_update(hgcd, q[0]); /* Recomputes u' and v' */
  assert(mpz_cmp(u, hgcd->R[2]) == 0);
  assert(mpz_cmp(v, hgcd->R[3]) == 0);

  mpz_swap(hgcd->a, r[0]);
  mpz_swap(hgcd->a, hgcd->b);

  /* Check r[1] */
  mpz_mul(u, q[1], hgcd->R[2]);
  mpz_add(u, u, hgcd->R[0]);

  mpz_mul(v, q[1], hgcd->R[3]);
  mpz_add(v, v, hgcd->R[1]);

  if (hgcd->sign >= 0)
    {
      /* r[0], u >= 0, r[1], v  <= 0 */
      if (mpz_cmp(r[1], v) < 0)
	goto done;
    }
  else
    {
      /* r[0], u <= 0, r[1], v >= 0 */
      if (mpz_cmp(r[1], u) < 0)
	goto done;
    }

  if (SIZE_IN_BITS(u) > l
      || SIZE_IN_BITS(v) > l
      || !jhgcd_condition(hgcd, u, v, r[1]))
    goto done;

  /* Ok, we can use r[1] too. */
  trace("jhgcd(%d): r[1] can be used.\n", count);
  jhgcd_update(hgcd, q[1]); /* Recomputes u' and v' */
  assert(mpz_cmp(u, hgcd->R[2]) == 0);
  assert(mpz_cmp(v, hgcd->R[3]) == 0); 

  mpz_swap(hgcd->a, r[1]);
  mpz_swap(hgcd->a, hgcd->b);
  
 done:
  jhgcd_clear(&R);
  jhgcd_clear(&S);
  mpz_clear(r[0]);
  mpz_clear(r[1]);
  mpz_clear(q[0]);
  mpz_clear(q[1]);
  mpz_clear(u);
  mpz_clear(v);

  return 1;
}
