#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <gmp.h>
#include <gmp-impl.h>
#include <longlong.h>

#include "trace.h"
#include "jebelean.h"

#if WANT_INLINE
# define INLINE inline
#else
# define INLINE
#endif

#define CHOOSE_M(n) (((n) + 1)/2)
#define CHOOSE_L(n) ((n)/2)

/* This should be tuned while looking at performance for large numbers */
#ifndef HGCD_LEHMER_THRESHOLD
# define HGCD_LEHMER_THRESHOLD 128
#endif

/* While this is important for small and medium size numbers. */
#ifndef FGCD_LEHMER_THRESHOLD
# define FGCD_LEHMER_THRESHOLD 128
#endif

/* Must have at least two limbs for hgcd to be defined */
#define HGCD_LIMIT 2

#define MPN_EQUAL_P(ap, asize, bp, bsize)			\
((asize) == (bsize) && !mpn_cmp((ap), (bp), (asize)))

#define MPN_LESS_P(ap, asize, bp, bsize)			\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) < 0))

#define MPN_LEQ_P(ap, asize, bp, bsize)				\
((asize) < (bsize) || ((asize) == (bsize)			\
                       && mpn_cmp((ap), (bp), (asize)) <= 0))

#define MPN_MUL_SIZE(asize, bsize) \
((asize) && (bsize) ? ((asize) + (bsize)) : 0)

/* Rotates pointers, (ap, bp, cp, dp)  <- (bp, cp, dp, ap) */
#define MP_PTR_SWAP4_LEFT(ap, bp, cp, dp)	\
do {						\
  mp_ptr __mp_ptr_swap_4_left_tmp = (ap);	\
  (ap) = (bp);					\
  (bp) = (cp);					\
  (cp) = (dp);					\
  (dp) = __mp_ptr_swap_4_left_tmp;		\
} while(0)

#define MP_PTR_SWAP4_RIGHT(ap, bp, cp, dp) \
MP_PTR_SWAP4_LEFT(dp, cp, bp, ap)

#define MPN_PTR_SHIFT4_LEFT(ap, asize, bp, bsize, cp, csize, dp, dsize)	\
do {									\
  MP_PTR_SWAP4_LEFT(ap, bp, cp, dp);					\
  (asize) = (bsize);							\
  (bsize) = (csize);							\
  (csize) = (dsize);							\
} while(0)
     
#define MP_PTR_SWAP3_LEFT(ap, bp, cp)	\
do {						\
  mp_ptr __mp_ptr_swap_3_left_tmp = (ap);	\
  (ap) = (bp);					\
  (bp) = (cp);					\
  (cp) = __mp_ptr_swap_3_left_tmp;		\
} while(0)

#define MPN_PTR_SHIFT3_LEFT(ap, asize, bp, bsize, cp, csize)	\
do {								\
  MP_PTR_SWAP3_LEFT(ap, bp, cp);				\
  (asize) = (bsize);						\
  (bsize) = (csize);						\
} while(0)

struct hgcd2
{
  mp_limb_t a[2];
  mp_limb_t b[2];
  int sign;
  mp_limb_t uv[4];
};

/* Almost like mpn_add, but without the requirements that s1size >=
   s2size. Returns size of the sum. */
static mp_limb_t
mpn_add_friendly(mp_ptr rp,
		 mp_srcptr s1p, mp_size_t s1size,
		 mp_srcptr s2p, mp_size_t s2size)
{
  mp_limb_t cy;
  
  ASSERT(s1size);
  ASSERT(s2size);
  
  if (s1size < s2size)
    MPN_SRCPTR_SWAP(s1p, s1size, s2p, s2size);

  cy = mpn_add(rp, s1p, s1size, s2p, s2size);
  if (cy)
    rp[s1size++] = cy;

  return s1size;
}

/* Almost like mpn_mul, but (i) without the requirement that s1size >=
   s2size, and (ii) returns "underflow indication": 1 iff the product
   was one limb shorter then the maximal size s1size + s2size. Input
   must still be non-zero. */
static mp_limb_t
mpn_mul_friendly(mp_ptr rp,
		 mp_srcptr s1p, mp_size_t s1size,
		 mp_srcptr s2p, mp_size_t s2size)
{
  ASSERT(s1size);
  ASSERT(s2size);
  
  if (s1size < s2size)
    MPN_SRCPTR_SWAP(s1p, s1size, s2p, s2size);

  return 0 == mpn_mul(rp, s1p, s1size, s2p, s2size);
}

/* Copied from mpn/generic/gcdext.c, and modified slightly to return
   the remainder. */
/* Two-limb division optimized for small quotients.  */
static INLINE mp_limb_t
div2 (mp_ptr rp, mp_srcptr np, mp_srcptr dp)
{
  mp_limb_t n0 = np[0];
  mp_limb_t n1 = np[1];
  mp_limb_t d0 = dp[0];
  mp_limb_t d1 = dp[1];

  mp_limb_t q = 0;
  
  if ((mp_limb_signed_t) n1 < 0)
    {
      int cnt;
      for (cnt = 1; (mp_limb_signed_t) d1 >= 0; cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      while (cnt)
	{
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  cnt--;
	}
    }
  else
    {
      int cnt;
      for (cnt = 0; n1 > d1 || (n1 == d1 && n0 >= d0); cnt++)
	{
	  d1 = (d1 << 1) | (d0 >> (GMP_LIMB_BITS - 1));
	  d0 = d0 << 1;
	}

      while (cnt)
	{
	  d0 = (d1 << (GMP_LIMB_BITS - 1)) | (d0 >> 1);
	  d1 = d1 >> 1;
	  q <<= 1;
	  if (n1 > d1 || (n1 == d1 && n0 >= d0))
	    {
	      sub_ddmmss (n1, n0, n1, n0, d1, d0);
	      q |= 1;
	    }
	  cnt--;
	}
    }
  
  rp[0] = n0;
  rp[1] = n1;
  return q;
}

/* Computes ua - vb, which must be non-negative.
   R == A is allowed, R == B is not.

   A and U must be non-zero.
   
   Result is stored at RP, of size at most ASIZE + 1. Returns size of
   the result. */

static mp_size_t
mpn_addsubmul_1(mp_ptr rp,
		mp_ptr ap, mp_size_t asize, mp_limb_t u, 
		mp_ptr bp, mp_size_t bsize, mp_limb_t v)
{
  mp_limb_t cy;

  ASSERT(asize);
  ASSERT(u);
  ASSERT(MPN_SAME_OR_SEPARATE_P(rp, ap, asize));
  ASSERT(!MPN_OVERLAP_P(ap, asize, bp, bsize));
  
  cy = mpn_mul_1(rp, ap, asize, u);
  if (cy)
    rp[asize++] = cy;

  if (v && bsize)
    {
      ASSERT(bsize <= asize);
      cy = mpn_submul_1(rp, bp, bsize, v);
      if (cy)
	{
	  ASSERT(bsize < asize);
	  cy = mpn_sub_1(rp + bsize, rp + bsize, asize - bsize, cy);
	}
      ASSERT(!cy);

      MPN_NORMALIZE(rp, asize);
    }
  return asize;
}

/* Compute au + bv, which must be non-zero */
static mp_size_t
mpn_addmul2_1(mp_ptr rp,
	      mp_ptr ap, mp_size_t asize, mp_limb_t u, 
	      mp_ptr bp, mp_size_t bsize, mp_limb_t v)
{
  mp_limb_t cy;
  mp_size_t size;

  if (v != 0 && (u == 0 || asize < bsize))
    {
      MPN_PTR_SWAP(ap, asize, bp, bsize);
      MP_LIMB_T_SWAP(u, v);
    }

  ASSERT(asize);
  
  size = asize;
  cy = mpn_mul_1(rp, ap, asize, u);
  if (cy)
    rp[size++] = cy;

  if (bsize != 0 && v != 0)
    {
      ASSERT(bsize <= size);
      cy = mpn_addmul_1(rp, bp, bsize, v);
      if (cy && bsize < size)
	cy = mpn_add_1(rp + bsize,
		       rp + bsize, size - bsize, cy);

      if (cy)
	rp[size++] = cy;
    }
  
  return size;
}


void
mpn_quotient_stack_init(struct mpn_quotient_stack *stack,
			mp_ptr limbs, mp_size_t alloc,
			mp_size_t *sizes, mp_size_t nsizes)
{
  stack->limbs = limbs;
  stack->alloc = alloc;
  stack->sizes = sizes;
  stack->nsizes = nsizes;

  stack->next_size = stack->sizes;
  stack->top = stack->limbs;

  stack->run = NULL;
}

mp_size_t
mpn_quotient_stack_limb_need(mp_size_t asize, mp_size_t bsize)
{
  /* We get the maximum number of limbs when we have the maximum
     number ow quotients. Any large, multi-limb quotient reduces the
     total number of limbs, because it replaces many small-limb
     quotients. */ 
  
  return 2 * asize * BITS_PER_MP_LIMB;
}

mp_size_t
mpn_quotient_stack_size_need(mp_size_t asize, mp_size_t bsize)
{
  /* The length of the quotient sequence is at most twice the number
     of bits in a. */
  return 2 * asize * BITS_PER_MP_LIMB;
}

void
trace_mpn_quotients(const char *label,
		    const struct mpn_quotient_stack *stack)
{
  mp_limb_t *p;
  mp_size_t *s;
  
  trace("%s: ", label);
  
  for (s = stack->sizes, p = stack->limbs; s < stack->next_size; s++)
    {
      mp_size_t size;

      ASSERT(p <= stack->top);
      size = *s;
      ASSERT(size);

      if (size < 0)
	trace("1 (%d times), ", -size);
      else
	{
	  trace("%Nd, ", p, size);
	  p += size;
	}
    }
  trace("\n");
}

/* To push a quotient, store it at stack->top, and call this
   function */
void
mpn_quotient_stack_push(struct mpn_quotient_stack *stack,
			mp_size_t qsize)
{
  ASSERT(stack->next_size < (stack->sizes + stack->nsizes));
  
  /* Handle the common case first */
  if (!qsize)
    {
      /* Run-length encoding... */
      if (stack->run)
	(*stack->run)--;
      
      else
	{
	  /* We need to start a new run */
	  stack->run = stack->next_size++;
	  *stack->run = -1;
	}
    }
  else
    {
      /* A large quotient */

      ASSERT(qsize);
      ASSERT(stack->top[qsize-1]);
      ASSERT(stack->top - stack->limbs + qsize <= stack->alloc);
      ASSERT(qsize != 1 || stack->top[0] > 1);

      *stack->next_size++ = qsize;
      stack->top += qsize;
      stack->run = NULL;
    }
}

static INLINE void
mpn_quotient_stack_push_1(struct mpn_quotient_stack *stack, mp_limb_t q)
{
  if (q != 1)
    {
      *stack->top++ = q;
      *stack->next_size++ = 1;
      stack->run = NULL;
    }
  else
    {
      /* Run-length encoding... */
      if (stack->run)
	(*stack->run)--;
      
      else
	{
	  /* We need to start a new run */
	  stack->run = stack->next_size++;
	  *stack->run = -1;
	}
    }
}

mp_size_t
mpn_quotient_stack_pop(struct mpn_quotient_stack *stack)
{
  ASSERT(stack->next_size > stack->sizes);
  if (stack->run)
    {
      (*stack->run)++;
      if (*stack->run == 0)
	{
	  stack->next_size--;
	  stack->run = NULL;
	}
      return 0;
    }
  else
    {
      mp_size_t qsize = stack->next_size[-1];

      ASSERT(qsize);
      if (qsize < 0)
	{
	  if (++stack->next_size[-1] == 0)
	    stack->next_size--;
	  else
	    stack->run = stack->next_size - 1;

	  return 0;
	}
      else
	{
	  stack->next_size--;
	  stack->top -= qsize;
	  ASSERT(qsize != 1 || stack->top[0] != 1);
	  
	  return qsize;
	}
    }
}

int
mpn_quotient_stack_equal(const struct mpn_quotient_stack *a,
			 const struct mpn_quotient_stack *b)
{
  return (a->next_size - a->sizes) == (b->next_size - b->sizes)
    && (a->top - a->limbs) == (b->top - b->limbs)
    && !memcmp(a->limbs, b->limbs,
	       (a->top - a->limbs) * sizeof(mp_limb_t))
    && !memcmp(a->sizes, b->sizes,
	       (a->next_size - a->sizes) * sizeof(mp_size_t));
}

static mp_size_t
mpn_jhgcd_tdiv(mp_ptr qp,
	       mp_ptr rp, mp_size_t *rsizep,
	       mp_srcptr ap, mp_size_t asize,
	       mp_srcptr bp, mp_size_t bsize)
{
  mp_size_t qsize = asize - bsize + 1;
  mp_size_t rsize = bsize;

  mpn_tdiv_qr(qp, rp, 0, ap, asize, bp, bsize);
  MPN_NORMALIZE(rp, rsize);
  *rsizep = rsize;

  if (qp[qsize - 1] == 0)
    qsize--;
  
  if (qsize == 1 && qp[0] == 1)
    return 0;

  return qsize;
}

mp_size_t
mpn_jhgcd_itch(mp_size_t asize, mp_size_t bsize)
{
  mp_size_t size = bsize + 1;
  mp_size_t alloc = CHOOSE_L(asize) + 1;

  return asize + 1 /* ap[0] */
    + size * 3  /* ap[1...3] */
    + alloc * 6;  /* uv array */
}

static void
mpn_jhgcd_identity(struct mpn_jhgcd *hgcd)
{
  hgcd->uvsize[0] = 1; hgcd->uvp[0][0] = 1;
  hgcd->uvsize[1] = 0;
  hgcd->uvsize[2] = 0;
  hgcd->uvsize[3] = 1; hgcd->uvp[3][0] = 1;
  
  hgcd->sign = 0;
}

/* Copies a and b into temporary storage. */
void
mpn_jhgcd_init(struct mpn_jhgcd *hgcd,
	       mp_srcptr ap, mp_size_t asize,
	       mp_srcptr bp, mp_size_t bsize,
	       mp_ptr tp, mp_size_t talloc)
{
  mp_size_t size = bsize + 1;
  mp_size_t alloc = CHOOSE_L(asize) + 1;
  unsigned i;

  ASSERT(asize + 1 + 3 * size <= talloc);
  hgcd->ap =  tp; tp += asize + 1, talloc -= asize + 1;
  hgcd->bp =  tp; tp += size, talloc -= size;
  hgcd->r0p = tp; tp += size; talloc -= size;
  hgcd->r1p = tp; tp += size; talloc -= size;

  MPN_COPY(hgcd->ap, ap, asize); hgcd->asize = asize;
  MPN_COPY(hgcd->bp, bp, bsize); hgcd->bsize = bsize;

  hgcd->alloc = alloc;
  ASSERT(6 * alloc <= talloc);
  for (i = 0; i<6; i++)
    {
      hgcd->uvp[i] = tp; tp += alloc; talloc -= alloc;
    }

  mpn_jhgcd_identity(hgcd);
}

void
trace_mpn_matrix(const char *label, const struct mpn_jhgcd *hgcd)
{
  trace("%s:\n"
	" %s(%Nd,\n"
	"   %Nd;\n"
	"   %Nd,\n"
	"   %Nd)\n",
	label, hgcd->sign < 0 ? "-" : " ",
	hgcd->uvp[0], hgcd->uvsize[0],
	hgcd->uvp[1], hgcd->uvsize[1],
	hgcd->uvp[2], hgcd->uvsize[2],
	hgcd->uvp[3], hgcd->uvsize[3]);
}

int
mpn_jhgcd_equal(const struct mpn_jhgcd *R,
		const struct mpn_jhgcd *S)
{
  unsigned i;
  if (!MPN_EQUAL_P(R->ap, R->asize, S->ap, S->asize))
    return 0;

  if (!MPN_EQUAL_P(R->bp, R->bsize, S->bp, S->bsize))
    return 0;

  if (R->sign != S->sign)
    return 0;

  for (i = 0; i<4; i++)
    if (!MPN_EQUAL_P(R->uvp[i], R->uvsize[i],
		     S->uvp[i], S->uvsize[i]))
      return 0;

  return 1;
}

/* Temporary space needed: hgcd->alloc + 1. FIXME: The +1 can probably
   be rationalized away, given that the uv array is allocated with one
   extra limb. */
#define MPN_JHGCD_UPDATE_ITCH(asize, bsize) ((asize) + 1)

static void
mpn_jhgcd_update(struct mpn_jhgcd *hgcd,
		 mp_srcptr qp, mp_size_t qsize,
		 /* Scratch space */
		 mp_ptr tp, mp_size_t talloc)
{
  unsigned i;

  /* Compute (u1, u2)  <-  (u2, u1 + q u2),
     by first computing u1 += q u2, then swapping. */
  
  if (qsize == 0)
    /* Represents a unit quotient */
    for (i = 0; i<2; i++)
      {
	mp_size_t u1size = hgcd->uvsize[i];
	mp_size_t u2size = hgcd->uvsize[i+2];
	mp_ptr u1p = hgcd->uvp[i];
	mp_ptr u2p = hgcd->uvp[i+2];

	ASSERT(hgcd->alloc >= u1size);
	ASSERT(hgcd->alloc >= u2size);

	if (u1size == 0)
	  {
	    MPN_COPY(u1p, u2p, u2size);
	    hgcd->uvsize[i] = u2size;	      
	  }
	else
	  {	      
	    if (u2size > 0)
	      {
		/* Compute u1 += u2, then swap */
		mp_limb_t cy;

		ASSERT(u2size >= u1size);
		  
		cy = mpn_add(u1p, u2p, u2size, u1p, u1size);
		u1size = u2size;
		if (cy)
		  {
		    ASSERT(u1size < hgcd->alloc);
		    u1p[u1size++] = cy;
		  }
	      }
	    /* SWAP */
	    hgcd->uvp[i] = u2p; hgcd->uvsize[i] = u2size;
	    hgcd->uvp[i+2] = u1p; hgcd->uvsize[i+2] = u1size;
	  }
      }
  else if (qsize == 1)
    {
      mp_limb_t q = qp[0];
      for (i = 0; i<2; i++)
	{
	  mp_size_t u1size = hgcd->uvsize[i];
	  mp_size_t u2size = hgcd->uvsize[i+2];

	  mp_ptr u1p = hgcd->uvp[i];
	  mp_ptr u2p = hgcd->uvp[i+2];
	  
	  mp_limb_t cy;
	  mp_size_t size;
      
	  ASSERT(hgcd->alloc >= u1size);
	  ASSERT(hgcd->alloc >= u2size);

	  if (u1size == 0)
	    {
	      /* Compute u1 = q u2 */
	      size = u2size;
	      cy = mpn_mul_1(u1p, u2p, u2size, q);
	      if (cy)
		{
		  ASSERT(size < hgcd->alloc);
		  u1p[size++] = cy;
		}
	    }
	  else if (u2size > 0)
	    {
	      ASSERT(u2size >= u1size);
	      size = u2size;
	      if (size > u1size)
		{
		  /* Do high limbs first */
		  cy = mpn_mul_1(u1p + u1size,
				 u2p + u1size, size - u1size, q);
		  if (cy)
		    {
		      ASSERT(size < hgcd->alloc);
		      u1p[size++] = cy;
		    }
		}
	      cy = mpn_addmul_1(u1p, u2p, u1size, q);
	      if (cy && size > u1size)
		cy = mpn_add_1(u1p + u1size,
			       u1p + u1size, size - u1size, cy);
	      if (cy)
		{
		  ASSERT(size < hgcd->alloc);
		  u1p[size++] = cy;
		}

	      hgcd->uvsize[i] = size;
	    }
	  else
	    /* u2 == 0, do nothing */
	    size = u1size;

	  MP_PTR_SWAP(hgcd->uvp[i], hgcd->uvp[i+2]);
	  hgcd->uvsize[i] = hgcd->uvsize[i+2];
	  hgcd->uvsize[i+2] = size;
	}
    }
  
  else for (i = 0; i<2; i++)
    {
      mp_size_t u1size = hgcd->uvsize[i];
      mp_size_t u2size = hgcd->uvsize[i+2];
      mp_ptr u1p = hgcd->uvp[i];
      mp_ptr u2p = hgcd->uvp[i+2];
      mp_size_t size;
      
      ASSERT(hgcd->alloc >= u1size);
      ASSERT(hgcd->alloc >= u2size);

      if (u1size == 0)
	{
	  size = qsize + u2size;
	  ASSERT(size <= hgcd->alloc);

	  size -= mpn_mul_friendly(u1p, u2p, u2size, qp, qsize);
	}
      else if (u2size > 0)
	{
	  mp_limb_t cy;
	  
	  size = qsize + u2size;
	  ASSERT(size <= hgcd->alloc);
	  ASSERT(size <= talloc);

	  /* This is the only place in this function where we need
	     to use the temporary storage TP. */
	  size -= mpn_mul_friendly(tp, u2p, u2size, qp, qsize);

	  ASSERT(size >= u1size);
	  cy = mpn_add(u1p, tp, size, u1p, u1size);
	  if (cy)
	    {
	      ASSERT(size < hgcd->alloc);
	      u1p[size++] = cy;
	    }
	}
      else
	/* Do nothing */
	size = u1size;
      
      MP_PTR_SWAP(hgcd->uvp[i], hgcd->uvp[i+2]);
      hgcd->uvsize[i] = hgcd->uvsize[i+2];
      hgcd->uvsize[i+2] = size;
    }

  hgcd->sign = ~hgcd->sign;

  ASSERT(hgcd->uvsize[0] <= hgcd->uvsize[2]);
  ASSERT(hgcd->uvsize[1] <= hgcd->uvsize[3]);
}

#define MPN_JHGCD_MUL_ITCH(asize, bsize) (3 * (CHOOSE_L(asize) + 1)

/* Multiply R by S from the left. */
static void
mpn_jhgcd_mul(const struct mpn_jhgcd *S, struct mpn_jhgcd *R,
	      mp_ptr tp, mp_size_t talloc)
{
  unsigned i;
  
  mp_size_t s0size = S->uvsize[0];
  mp_size_t s1size = S->uvsize[1];
  mp_size_t s2size = S->uvsize[2];
  mp_size_t s3size = S->uvsize[3];

  /* One column of R at a time */
  for (i = 0; i<2; i++)
    {
      mp_size_t r1size = R->uvsize[i];
      mp_size_t r2size = R->uvsize[i+2];
      mp_ptr r1p = R->uvp[i];
      mp_ptr r2p = R->uvp[i+2];
      mp_size_t t1size;
      mp_size_t t2size;
      mp_ptr t1p;
      mp_ptr t2p;

      /* The new value for a */
      mp_ptr np;
      mp_size_t nsize;
      
      /* Compute the new value for R[i], and put it in np, nsize */
      /* Allocate space, including one extra limb for the carry from
	 addition. Put the largest block first. */
      t1size = MPN_MUL_SIZE(s0size, r1size);
      t2size = MPN_MUL_SIZE(s1size, r2size);

      ASSERT(t1size < R->alloc);
      ASSERT(t2size < R->alloc);
      ASSERT(t1size + t2size + 1 < talloc);
      
      if (t1size >= t2size)
	{
	  t1p = np = tp;
	  t2p = t1p + t1size + 1;
	}
      else
	{
	  t2p = np = tp;
	  t1p = t2p + t2size + 1;
	}
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, S->uvp[0], s0size, r1p, r1size);
      if (t2size)
	t2size -= mpn_mul_friendly(t2p, S->uvp[1], s1size, r2p, r2size);

      if (t2size > t1size)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);

      nsize = t1size;
      if (t2size)
	{
	  mp_limb_t cy = mpn_add(np, t1p, t1size, t2p, t2size);
	  if (cy)
	    {
	      ASSERT(nsize < R->alloc);
	      np[nsize++] = cy;
	    }
	}
      else
	{
	  /* np should already have the largest number */
	  ASSERT(np == t1p);
	}

      ASSERT(nsize <= R->alloc);

      /* Compute the new value for the second element. We use
	 temporary space after np. */
      /* NOTE: I'd like to store t2p directly on top of r2p, but
	 mpn_mul doesn't allow source and destination to be the
	 same. */
      t1p = np + nsize;
      t1size = MPN_MUL_SIZE(s2size, r1size);
      t2p = t1p + t1size;
      t2size = MPN_MUL_SIZE(s3size, r2size);
      ASSERT(nsize + t1size + t2size <= talloc);

      ASSERT(t1size <= R->alloc);
      ASSERT(t2size <= R->alloc);
      
      if (t1size)
	t1size -= mpn_mul_friendly(t1p, S->uvp[2], s2size, r1p, r1size);
      if (t2size)
	t2size -= mpn_mul_friendly(t2p, S->uvp[3], s3size, r2p, r2size);

      if (t2size > t1size)
	MPN_PTR_SWAP(t1p, t1size, t2p, t2size);

      if (t2size)
	{
	  mp_limb_t cy = mpn_add(R->uvp[i+2], t1p, t1size, t2p, t2size);
	  if (cy)
	    {
	      ASSERT(t1size < R->alloc);
	      R->uvp[i+2][t1size++] = cy;
	    }
	}
      else
	{
	  MPN_COPY(R->uvp[i+2], t1p, t1size);
	}
      
      /* Now the second element is stored at the right place, and the
	 first is in np, nsize. So copy the latter, and update both
	 signs. */

      MPN_COPY(R->uvp[i], np, nsize);
      R->uvsize[i] = nsize;
      R->uvsize[i+2] = t1size;
    }
  R->sign ^= S->sign;
}

/* Multiply R by S from the left. */
static void
mpn_jhgcd_mul_1(const struct hgcd2 *S, struct mpn_jhgcd *R)
{
  /* Use R->uvsize[4] and R->uv[5] for temporary storage */
  unsigned i;
  
  /* One column of R at a time */
  for (i = 0; i<2; i++)
    {
      R->uvsize[4] = mpn_addmul2_1(R->uvp[4],
				   R->uvp[i], R->uvsize[i], S->uv[0],
				   R->uvp[i+2], R->uvsize[i+2], S->uv[1]);
      ASSERT(R->uvsize[4] <= R->alloc);

      R->uvsize[5] = mpn_addmul2_1(R->uvp[5],
				   R->uvp[i], R->uvsize[i], S->uv[2],
				   R->uvp[i+2], R->uvsize[i+2], S->uv[3]);
      ASSERT(R->uvsize[5] <= R->alloc);
      
				   
      MPN_PTR_SWAP(R->uvp[i], R->uvsize[i],
		   R->uvp[4], R->uvsize[4]);
      MPN_PTR_SWAP(R->uvp[i+2], R->uvsize[i+2],
		   R->uvp[5], R->uvsize[5]);
    }
  R->sign ^= S->sign;
}

#define JHGCD_BACKUP_ITCH(asize, bsize) (2 * (asize))

/* Returns qsize. Temporary space needed is hgcd->alloc + qsize */
static mp_size_t
jhgcd_backup(struct mpn_jhgcd *hgcd,
	     struct mpn_quotient_stack *stack,
	     mp_ptr tp, mp_size_t talloc)
{
  mp_size_t qsize;
  mp_limb_t cy;
  unsigned i;

  MP_PTR_SWAP4_RIGHT(hgcd->ap, hgcd->bp, hgcd->r0p, hgcd->r1p);
  hgcd->r1size = hgcd->r0size;
  hgcd->r0size = hgcd->bsize;
  hgcd->bsize  = hgcd->asize;
  
  qsize = mpn_quotient_stack_pop(stack);

  if (!qsize)
    {
      /* Compute a_{i-1} = a_i + a_{i+1} */
      ASSERT(hgcd->bsize >= hgcd->r0size);

      hgcd->asize = hgcd->bsize;
      cy = mpn_add(hgcd->ap, hgcd->bp, hgcd->bsize, hgcd->r0p, hgcd->r0size);
      if (cy)
	hgcd->ap[hgcd->asize++] = cy;

      /* Same for the R matrix, one column at a time. But as we're working
	 with absolute values, we get a subtraction. */
      for (i = 0; i<2; i++)
	{
	  mp_size_t s = hgcd->uvsize[i+2];

	  ASSERT(hgcd->uvsize[i] <= s);

	  cy = mpn_sub(hgcd->uvp[i+2],
		       hgcd->uvp[i+2], s,
		       hgcd->uvp[i], hgcd->uvsize[i]);
	  ASSERT(!cy);
	  MPN_NORMALIZE(hgcd->uvp[i+2], s);
	  
	  MP_PTR_SWAP(hgcd->uvp[i], hgcd->uvp[i+2]);
	  hgcd->uvsize[i+2] = hgcd->uvsize[i];
	  hgcd->uvsize[i] = s;
	}
    }
  else
    {
      mp_srcptr qp = stack->top;

      /* FIXME: Special case for qsize = 1? */
      /* Compute a_{i-1} = q a_i + a_{i+1} */

      /* Example: a1 = 8, a2 = 5, q = 1
	 a0 <- q a1 + a2 = 13 */

      hgcd->asize = MPN_MUL_SIZE(hgcd->bsize, qsize);
      ASSERT(hgcd->asize);
  
      hgcd->asize -= mpn_mul_friendly(hgcd->ap,
				      hgcd->bp, hgcd->bsize, qp, qsize);

      ASSERT(hgcd->asize >= hgcd->r0size);
      cy = mpn_add(hgcd->ap, hgcd->ap, hgcd->asize, hgcd->r0p, hgcd->r0size);
      if (cy)
	hgcd->ap[hgcd->asize++] = cy;
  
      /* Same for the R matrix, one column at a time. But as we're working
	 with absolute values, we get a subtraction. */
      for (i = 0; i<2; i++)
	{
	  mp_size_t tsize = MPN_MUL_SIZE(hgcd->uvsize[i], qsize);
	  ASSERT(tsize <= talloc);
      
	  if (tsize)
	    {		
	      tsize -= mpn_mul_friendly(tp,
					hgcd->uvp[i], hgcd->uvsize[i],
					qp, qsize);
	      ASSERT(tsize <= hgcd->uvsize[i+2]);
	      ASSERT(tsize);

	      trace("  i = %d, t = %Nd\n",
		    i, tp, tsize);
	  
	      cy = mpn_sub(hgcd->uvp[i+2],
			   hgcd->uvp[i+2], hgcd->uvsize[i+2],
			   tp, tsize);
	      ASSERT(!cy);
	      MPN_NORMALIZE(hgcd->uvp[i+2], hgcd->uvsize[i+2]);
	    }
      
	  MPN_PTR_SWAP(hgcd->uvp[i], hgcd->uvsize[i],
		       hgcd->uvp[i+2], hgcd->uvsize[i+2]);
	}
    }
  hgcd->sign = ~hgcd->sign;

  return qsize;
}

/* Computes (u a + c)/b, where there must be no remainder, and a one-limb
   quotient. */
static mp_limb_t
compute_vplus(mp_srcptr a, mp_srcptr b, mp_srcptr c, mp_limb_t u)
{
  mp_limb_t t[3];
  mp_limb_t r[2];
  mp_limb_t q[2];
  
  mp_size_t tsize;
  mp_limb_t cy;

  ASSERT(a[1]);
  ASSERT(b[1]);
  
  MPN_COPY(t, c, 2);
  tsize = 2;
  cy = mpn_addmul_1(t, a, 2, u);
  if (cy)
    t[tsize++] = cy;

  else
    {
      /* We must have t >= b */
      ASSERT(t[1] > b[1]
	     || (t[1] == b[1] && t[0] >= b[0]));
    }      
  
  q[1] = 0;
  mpn_tdiv_qr(q, r, 0, t, tsize, b, 2);
  ASSERT(r[0] == 0 && r[1] == 0);
  ASSERT(q[1] == 0);

  return q[0];
}

/* Computes (u a - c) / b, where there must be no remainder, and a
   one-limb quotient. */
static mp_limb_t
compute_vminus(mp_srcptr a, mp_srcptr b, mp_srcptr c, mp_limb_t u)
{
  mp_limb_t t[3];
  mp_limb_t r[2];
  mp_limb_t q[2];
  
  mp_size_t tsize;
  mp_limb_t cy;

  ASSERT(a[1]);
  ASSERT(b[1]);

  tsize = 2;
  cy = mpn_mul_1(t, a, 2, u);
  if (cy)
    t[tsize++] = cy;

  cy = mpn_sub(t, t, tsize, c, 2);
  ASSERT(!cy);
  MPN_NORMALIZE(t, tsize);

  /* We must have t == 0 or t >= b */
  if (tsize == 0)
    return 0;

  ASSERT(tsize >= 2);
  q[1] = 0;
  mpn_tdiv_qr(q, r, 0, t, tsize, b, 2);
  
  ASSERT(r[0] == 0 && r[1] == 0);
  ASSERT(q[1] == 0);
  
  return q[0];
}

static int
jhgcd2_check_q(struct hgcd2 *hgcd, mp_srcptr b, mp_limb_t q, mp_srcptr r)
{
  mp_limb_t uv[2];
  mp_limb_t t[2];
  mp_limb_t s[2];

  unsigned i;
  
  /* Compute next row of the uv matrix. There should be no overflow */
  uv[0] = q * hgcd->uv[2] + hgcd->uv[0];
  uv[1] = q * hgcd->uv[3] + hgcd->uv[1];

  if (hgcd->sign < 0)
    {
      if (r[1] == 0 && r[0] < uv[0])
	return 0;
      
      /* Check if a_{k+1} - a{k+2} >= v_{k+2} - v{k+1} (k+1 even)
	 where

	 a_{k+1} = hgcd->b,
	 a_{k+2} = r,
	 v_{k+1} = -hgcd->uv[3],
	 v_{k+2} = v_k - q v_{k+1}
	         = hgcd->uv[1] + q * hgcd->uv[3])
		 = v
      */
      s[0] = hgcd->uv[3] + uv[1];
      s[1] = (s[0] < uv[1]);
    }
  else
    {
      if (r[1] == 0 && r[0] < uv[1])
	return 0;
      
      /* Check if a_{k+1} - a{k+2} >= u_{k+2} - u{k+1} (k+1 odd)
	 where
	 
	 a_{k+1} = hgcd->b,
	 a_{k+2} = r,
	 u_{k+1} = -hgcd->uv[2],
	 u_{k+2} = u_k - q[0] u_{k+1}
	         = hgcd->uv[0] + q * hgcd->uv[2])
		 = u
      */
      s[0] = hgcd->uv[2] + uv[0]; 
      s[1] = (s[0] < uv[0]);
    }

  sub_ddmmss(t[1], t[0], b[1], b[0], r[1], r[0]);
  
  if (t[1] < s[1] || (t[1] == s[1] && t[0] < s[0]))
    return 0;
  
  for (i = 0; i<2; i++)
    {
      MP_LIMB_T_SWAP(uv[i], hgcd->uv[i]);
      MP_LIMB_T_SWAP(hgcd->uv[i], hgcd->uv[i+2]);
    }

  hgcd->sign = ~hgcd->sign;
  return 1;
}

static int
hgcd2(struct hgcd2 *hgcd, struct mpn_quotient_stack *quotients)
{
  mp_limb_t limbs[8];
  mp_ptr ap = limbs;
  mp_ptr bp = limbs + 2;
  mp_ptr r0 = limbs + 4;
  mp_ptr r1 = limbs + 6;
  mp_limb_t q0;
  mp_limb_t q1;
  
  mp_limb_t u0 = 1;
  mp_limb_t u1 = 0;
#if WANT_ASSERT
  mp_limb_t v0 = 0;
  mp_limb_t v1 = 1;
#endif
  int sign = 0;

  ASSERT(hgcd->a[1] > hgcd->b[1]
	 || (hgcd->a[1] == hgcd->b[1] && hgcd->a[0] >= hgcd->b[0]));
  
  MPN_COPY(ap, hgcd->a, 2);
  MPN_COPY(bp, hgcd->b, 2);
  
  q0 = div2(r0, ap, bp);
  if (r0[1] == 0)
    return 0;

  for (;;)
    {
      mp_limb_t t;

      q1 = div2(r1, bp, r0);
      if (r1[1] == 0)
	break;

      if (quotients)
	mpn_quotient_stack_push_1(quotients, q0);
      
      MP_PTR_SWAP(r1, ap);
      MP_PTR_SWAP(ap, bp);
      MP_PTR_SWAP(bp, r0);

      t = u0 + q0 * u1;
      u0 = u1; u1 = t;

#if WANT_ASSERT
      t = v0 + q0 * v1;
      v0 = v1; v1 = t;
#endif
      
      q0 = q1;
      sign = ~sign;
    }
  
  ASSERT(ap[1] != 0);
  ASSERT(bp[1] != 0);
  ASSERT(r0[1] != 0);
  ASSERT(r1[1] == 0);

  /* The uv matrix is (u0,v0 ; u1,v1), where we have already computed
     u0 and u1. From ap = u0 a + v0 b we get v0 = (ap - u0 a) / b, and
     similarly v1 = (bp - u1 a) / b. */
  
  hgcd->uv[0] = u0;
  hgcd->uv[2] = u1;

  hgcd->sign = sign;

  /*   u a - v b = c   ==>  v = (u a - c) / b
     - u a + v b = c   ==>  v = (u a + c) / b */
  if (sign >= 0)
    {
      hgcd->uv[1] = compute_vminus(hgcd->a, hgcd->b, ap, u0);
      hgcd->uv[3] = compute_vplus(hgcd->a, hgcd->b, bp, u1);
    }
  else
    {
      hgcd->uv[1] = compute_vplus(hgcd->a, hgcd->b, ap, u0);
      hgcd->uv[3] = compute_vminus(hgcd->a, hgcd->b, bp, u1);
    }

  ASSERT(hgcd->uv[1] == v0);
  ASSERT(hgcd->uv[3] == v1);
  
  /* Check r0 */
  if (!jhgcd2_check_q(hgcd, bp, q0, r0))
    return 1;

  /* We can use r0 */
  if (quotients)
    mpn_quotient_stack_push_1(quotients, q0);

  MP_PTR_SWAP(r0, ap);
  MP_PTR_SWAP(ap, bp);
  
  if (!jhgcd2_check_q(hgcd, bp, q1, r1))
    return 1;

  /* We can use r1 */  
  if (quotients)
    mpn_quotient_stack_push_1(quotients, q1);

  MP_PTR_SWAP(r1, ap);
  MP_PTR_SWAP(ap, bp);
  
  return 1;
}


#define SHIFT_LEFT(count, x) \
  (((x)[1] << (count)) | ((x)[0] >> (BITS_PER_MP_LIMB - (count))))

/* lehmer varient using Jebelean's ideas */
/* tp should point an area of at least asize + 1 limbs */
mp_size_t
mpn_gcd_jebelean(mp_ptr rp,
		 mp_ptr ap, mp_size_t asize,
		 mp_ptr bp, mp_size_t bsize,
		 mp_ptr tp)
{
  mp_size_t rsize;
  mp_size_t tsize;
  
  ASSERT(asize >= bsize);
  ASSERT(bsize);
  ASSERT(ap[asize - 1]);
  ASSERT(bp[bsize - 1]);

  /* Start with one long division */  
  tsize = mpn_jhgcd_tdiv(tp, rp, &rsize,
			 ap, asize, bp, bsize);

  if (!rsize)
    {
      MPN_COPY(rp, bp, bsize);
      return bsize;
    }
  else
    {
      MP_PTR_SWAP(ap, bp);
      asize = bsize;
      MPN_COPY(bp, rp, rsize);
      bsize = rsize;
    }

  while (bsize >= 2)
    {
      struct hgcd2 hgcd;

      /* The case asize == 2 is needed to take care of values that are
	 between one and two *full* limbs in size. */
      if (asize == 2 || (ap[asize-1] & GMP_LIMB_HIGHBIT))
	{
	  if (bsize < asize)
	    goto euclid;
	  
	  hgcd.a[0] = ap[asize - 2];
	  hgcd.a[1] = ap[asize - 1];

	  ASSERT(asize == bsize);
	  hgcd.b[0] = bp[asize - 2];
	  hgcd.b[1] = bp[asize - 1];
	}
      else
	{
	  unsigned shift;
	  if (bsize + 1 < asize)
	    goto euclid;

	  /* We want two *full* limbs */
	  ASSERT(asize > 2);

	  count_leading_zeros(shift, ap[asize-1]);
	  if (bsize == asize)
	    hgcd.b[1] = SHIFT_LEFT(shift, bp + asize - 2);
	  else
	    {
	      ASSERT(asize == bsize + 1);
	      hgcd.b[1] = bp[asize - 2] >> (BITS_PER_MP_LIMB - shift);
	    }
	  if (!hgcd.b[1])
	    goto euclid;
	  hgcd.b[0] = SHIFT_LEFT(shift, bp + asize - 3);
	  
	  hgcd.a[0] = SHIFT_LEFT(shift, ap + asize - 3);
	  hgcd.a[1] = SHIFT_LEFT(shift, ap + asize - 2);
	}

      if (hgcd2(&hgcd, NULL))
	{
	  mp_size_t tsize;

	  /* uv[0] may be zero, all other must be non-zero */
	  ASSERT(hgcd.uv[1]);
	  ASSERT(hgcd.uv[2]);
	  ASSERT(hgcd.uv[3]);

	  if (hgcd.sign >= 0)
	    {
	      /* a <- u a - v b
		 b <- v'b - u'a */

	      ASSERT(hgcd.uv[0]);

	      /* Compute the new value for b first, storing it in
		 tp */
	      tsize = mpn_addsubmul_1(tp,
				      bp, bsize, hgcd.uv[3],
				      ap, asize, hgcd.uv[2]);

	      /* Compute new value for a */
	      asize = mpn_addsubmul_1(ap,
				      ap, asize, hgcd.uv[0],
				      bp, bsize, hgcd.uv[1]);

	      MP_PTR_SWAP(bp, tp);
	      bsize = tsize;
	    }
	  else
	    {
	      /* a <- v b - u a
		 b <- u'a - v'b */

	      /* Compute new value for b, storing it in tp */
	      tsize = mpn_addsubmul_1(tp,
				      ap, asize, hgcd.uv[2], 
				      bp, bsize, hgcd.uv[3]);
	      /* Compute new value for a, storing it in bp */
	      bsize = mpn_addsubmul_1(bp,
				      bp, bsize, hgcd.uv[1], 
				      ap, asize, hgcd.uv[0]);

	      MP_PTR_SWAP(ap, tp);
	      MP_PTR_SWAP(ap, bp);
	      asize = bsize;
	      bsize = tsize;
	    }
	  ASSERT(MPN_LESS_P(bp, bsize, ap, asize));
	}
      else
	{
	euclid:
	  tsize = mpn_jhgcd_tdiv(tp, rp, &rsize,
				 ap, asize, bp, bsize);
	  
	  if (!rsize)
	    {
	      MPN_COPY(rp, bp, bsize);
	      return bsize;
	    }

	  MPN_COPY(ap, rp, rsize);
	  MP_PTR_SWAP(ap, bp);
	  asize = bsize; bsize = rsize;
	}
    }

  {
    mp_limb_t a;
    mp_limb_t b;
    unsigned k;
    
    /* Bitwise gcd for the rest */
    ASSERT(bsize == 1);
    b = bp[0];
    
    if (asize == 1)
      a = ap[0];
    else
      a = mpn_divmod_1(ap, ap, asize, b);

    for (k = 0; ( (a & 1) == 0
		  && (b & 1) == 0 );)
      {
	a /= 2; b /= 2;
	k++;
      }

    /* Make sure a is odd */
    if ( (a & 1) == 0)
      MP_LIMB_T_SWAP(a, b);

    ASSERT( (a & 1) == 1);

    while (b)
      {
	while ( (b & 1) == 0)
	  b /= 2;

	if (b > a)
	  b -= a;
	else
	  {
	    mp_limb_t t = a - b;
	    a = b; b = t;
	  }
      }

    rp[0] = a << k;
    return 1;
  }
}

#define JHGCD_CHECK_Q_ITCH(asize, bsize) (bsize)

static int
jhgcd_check_q(struct mpn_jhgcd *hgcd, mp_srcptr qp, mp_size_t qsize,
	      mp_ptr tp, mp_size_t talloc)
{
  mp_size_t tsize;
  mp_limb_t cy;
  unsigned i;

  /* Compute next row of the uv matrix,
     uv[4+i] = uv[i] + q uv[2 + i]
     
     FIXME: Almost duplicates code in mpn_jhgcd_update. */
  
  if (qsize == 0)
    for (i = 0; i<2; i++)
      {
	mp_size_t u1size = hgcd->uvsize[i];
	mp_size_t u2size = hgcd->uvsize[i+2];
	mp_ptr u1p = hgcd->uvp[i];
	mp_ptr u2p = hgcd->uvp[i+2];
	mp_ptr u3p = hgcd->uvp[i+4];

	if (u1size == 0)
	  {
	    MPN_COPY(u3p, u2p, u2size);
	    hgcd->uvsize[i+4] = u2size;
	  }
	else if (u2size == 0)
	  {
	    MPN_COPY(u3p, u1p, u1size);
	    hgcd->uvsize[i+4] = u1size;
	  }
	
	else
	  {
	    mp_size_t size = u2size;
	    mp_limb_t cy;

	    ASSERT(size >= u1size);
	    
	    cy = mpn_add(u3p, u2p, u2size, u1p, u1size);
	    if (cy)
	      {
		ASSERT(size < hgcd->alloc);
		u3p[size++] = cy;
	      }
	    hgcd->uvsize[i+4] = size;
	  }
      }

  else if (qsize == 1)
    {
      mp_limb_t q = qp[0];
      for (i = 0; i<2; i++)
	{
	  mp_size_t u1size = hgcd->uvsize[i];
	  mp_size_t u2size = hgcd->uvsize[i+2];
	  mp_ptr u1p = hgcd->uvp[i];
	  mp_ptr u2p = hgcd->uvp[i+2];
	  mp_ptr u3p = hgcd->uvp[i+4];
	  
	  if (u2size == 0)
	    {
	      MPN_COPY(u3p, u1p, u1size);
	      hgcd->uvsize[i+4] = u1size;
	    }
	  else
	    {
	      mp_size_t size;
	      mp_limb_t cy;
	      
	      ASSERT(u2size >= u1size);
	      size = u2size;
	      cy = mpn_mul_1(u3p, u2p, u2size, q);
	      if (cy)
		{
		  ASSERT(size < hgcd->alloc);
		  u3p[size++] = cy;
		}

	      if (u1size)
		{
		  cy = mpn_add(u3p, u3p, size, u1p, u1size);
		  if (cy)
		    {
		      ASSERT(size < hgcd->alloc);
		      u3p[size++] = cy;
		    }
		}
	      hgcd->uvsize[i+4] = size;
	    }
	}
    }
  else for (i = 0; i<2; i++)
    {
      mp_size_t u1size = hgcd->uvsize[i];
      mp_size_t u2size = hgcd->uvsize[i+2];
      mp_ptr u1p = hgcd->uvp[i];
      mp_ptr u2p = hgcd->uvp[i+2];
      mp_ptr u3p = hgcd->uvp[i+4];

      if (u2size == 0)
	{
	  MPN_COPY(u3p, u1p, u1size);
	  hgcd->uvsize[i+4] = u1size;
	}
      else
	{
	  mp_size_t size;
	  
	  size = u2size + qsize;
	  ASSERT(size <= hgcd->alloc);
	  
	  size -= mpn_mul_friendly(u3p, u2p, u2size, qp, qsize);

	  if (u1size)
	    {
	      mp_limb_t cy;
	      cy = mpn_add(u3p, u3p, size, u1p, u1size);
	      if (cy)
		{
		  ASSERT(size < hgcd->alloc);
		  u3p[size++] = cy;
		}
	    }
	  hgcd->uvsize[i+4] = size;
	}
    }

  tsize = hgcd->bsize;
  ASSERT(tsize <= talloc);
  
  cy = mpn_sub(tp, hgcd->bp, tsize, hgcd->r0p, hgcd->r0size);
  ASSERT(!cy);

  MPN_NORMALIZE(tp, tsize);
  trace("jhgcd_check_q: b - r0 = %Nd\n",
	tp, tsize);
  
  if (hgcd->sign < 0)
    {
      if (MPN_LESS_P(hgcd->r0p, hgcd->r0size, hgcd->uvp[4], hgcd->uvsize[4]))
	return 0;

      /* Check t >= uv[5] + uv[3] */
      i = 1;
    }
  else
    {
      if (MPN_LESS_P(hgcd->r0p, hgcd->r0size, hgcd->uvp[5], hgcd->uvsize[5]))
	return 0;

      /* t >= Check uv[4] + uv[2] */
      i = 0;
    }
  if (tsize < hgcd->uvsize[2+i]
      || mpn_sub(tp, tp, tsize, hgcd->uvp[2+i], hgcd->uvsize[2+i]))
    return 0;
  
  MPN_NORMALIZE(tp, tsize);
  
  if (MPN_LESS_P(tp, tsize, hgcd->uvp[4 + i], hgcd->uvsize[4 + i]))
    return 0;

  /* The remainder can be used. Update. */
  MPN_PTR_SHIFT4_LEFT(hgcd->ap,  hgcd->asize,  hgcd->bp,  hgcd->bsize,
		      hgcd->r0p, hgcd->r0size, hgcd->r1p, hgcd->r1size);

  MPN_PTR_SHIFT3_LEFT(hgcd->uvp[0], hgcd->uvsize[0], 
		      hgcd->uvp[2], hgcd->uvsize[2], 
		      hgcd->uvp[4], hgcd->uvsize[4]);
  MPN_PTR_SHIFT3_LEFT(hgcd->uvp[1], hgcd->uvsize[1], 
		      hgcd->uvp[3], hgcd->uvsize[3], 
		      hgcd->uvp[5], hgcd->uvsize[5]);

  hgcd->sign = ~hgcd->sign;

  return 1;
}

mp_size_t
mpn_jhgcd_slow_itch(struct mpn_jhgcd *hgcd)
{
  return 2 * (hgcd->asize + 1)
    + hgcd->alloc + 1;
}

int
mpn_jhgcd_slow(struct mpn_jhgcd *hgcd,
	       struct mpn_quotient_stack *quotients,
	       mp_ptr tp, mp_size_t talloc)
{
  mp_size_t n = hgcd->asize;
  mp_size_t m = CHOOSE_M(n);
  mp_size_t l = CHOOSE_L(n);
  
  mp_size_t q0size;
  mp_size_t q1size;
  
  unsigned i;
  
  ASSERT(m);

  if (hgcd->bsize <= m)
    return 0;
    
  /* Maintain r[0], q[0] quotient and remainder of a and b, r[1], q[1]
   * quotient and remainder of b and r[0].
   *
   * We should have a, b, r[0] >= W^m (or #a > m etc), and loop until
   * r[1] < 2^m (or #r[1] <= m). */

  /* Needed for the mpn_jhgcd_update call */
  ASSERT(hgcd->alloc + 1 <= talloc);

  q0size = mpn_jhgcd_tdiv(quotients->top, hgcd->r0p, &hgcd->r0size,
			  hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  if (hgcd->r0size <= m)
    return 0;
  
  for (;;)
    {      
      ASSERT(MPN_LESS_P(hgcd->bp, hgcd->bsize, hgcd->ap, hgcd->asize));
      ASSERT(MPN_LESS_P(hgcd->r0p, hgcd->r0size, hgcd->bp, hgcd->bsize));

      q1size = mpn_jhgcd_tdiv(quotients->top + q0size,
			      hgcd->r1p, &hgcd->r1size,
			      hgcd->bp, hgcd->bsize,
			      hgcd->r0p, hgcd->r0size);
            
      ASSERT(MPN_LESS_P(hgcd->r1p, hgcd->r1size, hgcd->r0p, hgcd->r0size));
      
      if (hgcd->r1size <= m)
	break;

      mpn_jhgcd_update(hgcd, quotients->top, q0size, tp, talloc);
      mpn_quotient_stack_push(quotients, q0size);

      /* Rotate remainders */
      MPN_PTR_SHIFT4_LEFT(hgcd->ap,  hgcd->asize,  hgcd->bp,  hgcd->bsize,
			  hgcd->r0p, hgcd->r0size, hgcd->r1p, hgcd->r1size);
      
      q0size = q1size;
    }
  
  /* Now a,b,r[0] >= W^m, r[1] < W^m */
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);
  ASSERT(hgcd->r0size > m);
  ASSERT(hgcd->r1size <= m);

#if WANT_ASSERT
  for (i = 0; i<4; i++)
    ASSERT(hgcd->uvsize[i] <= l);
#endif

  if (!jhgcd_check_q(hgcd, quotients->top, q0size, tp, talloc))
    return 1;

  /* We can use r0 */
  trace("mpn_jhgcd_slow: r0 can be used.\n");

  mpn_quotient_stack_push(quotients, q0size);

  if (!jhgcd_check_q(hgcd, quotients->top, q1size, tp, talloc))
    return 1;

  /* We can use r1 */
  trace("mpn_jhgcd_slow: r1 can be used.\n");
  mpn_quotient_stack_push(quotients, q1size);

  return 1;
}

static int
euclid_step(struct mpn_jhgcd *hgcd,
	    struct mpn_quotient_stack *quotients,
	    mp_size_t m,
	    mp_ptr tp, mp_size_t talloc)
{
  mp_size_t qsize;

  qsize = mpn_jhgcd_tdiv(quotients->top, hgcd->r0p, &hgcd->r0size,
			 hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  if (hgcd->r0size <= m)
    return 0;
  
  MPN_PTR_SHIFT4_LEFT(hgcd->ap,  hgcd->asize,  hgcd->bp,  hgcd->bsize,
		      hgcd->r0p, hgcd->r0size, hgcd->r1p, hgcd->r1size);

  mpn_jhgcd_update(hgcd, quotients->top, qsize,
		   tp, talloc);
  mpn_quotient_stack_push(quotients, qsize);

  return 1;
}

static int
lehmer_step(struct mpn_jhgcd *hgcd,
	    struct mpn_quotient_stack *quotients)
{
  struct hgcd2 h2;
  if (hgcd->bsize < 2)
    return 0;

  /* The case asize == 2 is needed to take care of values that are
     between one and two *full* limbs in size. */
  if (hgcd->asize == 2 || (hgcd->ap[hgcd->asize-1] & GMP_LIMB_HIGHBIT))
    {
      if (hgcd->bsize < hgcd->asize)
	return 0;
	  
      h2.a[0] = hgcd->ap[hgcd->asize - 2];
      h2.a[1] = hgcd->ap[hgcd->asize - 1];

      ASSERT(hgcd->asize == hgcd->bsize);
      h2.b[0] = hgcd->bp[hgcd->asize - 2];
      h2.b[1] = hgcd->bp[hgcd->asize - 1];
    }
  else
    {
      unsigned shift;
      if (hgcd->bsize + 1 < hgcd->asize)
	return 0;
      
      /* We want two *full* limbs */
      ASSERT(hgcd->asize > 2);

      count_leading_zeros(shift, hgcd->ap[hgcd->asize-1]);
      if (hgcd->bsize == hgcd->asize)
	h2.b[1] = SHIFT_LEFT(shift, hgcd->bp + hgcd->asize - 2);
      else
	{
	  ASSERT(hgcd->asize == hgcd->bsize + 1);
	  h2.b[1] = hgcd->bp[hgcd->asize - 2] >> (BITS_PER_MP_LIMB - shift);
	}
      if (!h2.b[1])
	return 0;
      
      h2.b[0] = SHIFT_LEFT(shift, hgcd->bp + hgcd->asize - 3);
	  
      h2.a[0] = SHIFT_LEFT(shift, hgcd->ap + hgcd->asize - 3);
      h2.a[1] = SHIFT_LEFT(shift, hgcd->ap + hgcd->asize - 2);
    }

  if (!hgcd2(&h2, quotients))
    return 0;
  
  /* uv[0] may be zero, all other must be non-zero */
  ASSERT(h2.uv[1]);
  ASSERT(h2.uv[2]);
  ASSERT(h2.uv[3]);

  /* Compute new values for a and b, rotating the A
     array to get space. */

  if (h2.sign >= 0)
    {
      /* a <- u a - v b
	 b <- v'b - u'a */

      ASSERT(h2.uv[0]);

      hgcd->r0size = mpn_addsubmul_1(hgcd->r0p,
			       hgcd->ap, hgcd->asize, h2.uv[0],
			       hgcd->bp, hgcd->bsize, h2.uv[1]);
      hgcd->r1size = mpn_addsubmul_1(hgcd->r1p, 
			       hgcd->bp, hgcd->bsize, h2.uv[3],
			       hgcd->ap, hgcd->asize, h2.uv[2]);
    }
  else
    {
      /* a <- v b - u a
	 b <- u'a - v'b */

      hgcd->r0size = mpn_addsubmul_1(hgcd->r0p,
			       hgcd->bp, hgcd->bsize, h2.uv[1],
			       hgcd->ap, hgcd->asize, h2.uv[0]);

      hgcd->r1size = mpn_addsubmul_1(hgcd->r1p,
			       hgcd->ap, hgcd->asize, h2.uv[2],
			       hgcd->bp, hgcd->bsize, h2.uv[3]);
    }

  MP_PTR_SWAP(hgcd->ap, hgcd->r0p);
  MP_PTR_SWAP(hgcd->bp, hgcd->r1p);
  hgcd->asize = hgcd->r0size;
  hgcd->bsize = hgcd->r1size;
  
  ASSERT(MPN_LESS_P(hgcd->bp, hgcd->bsize, hgcd->ap, hgcd->asize));

  /* Update the matrix */
  mpn_jhgcd_mul_1(&h2, hgcd);

  return 1;
}

/* Loops using lehmer steps, and returns values satisfying the hgcd
   requirements. Might pop one value from the stack, so it it can't be
   called with an empty stack and #A > m, #B > m and #(A % B) <= m. */
static void
lehmer_loop(struct mpn_jhgcd *hgcd,
	    struct mpn_quotient_stack *quotients,
	    mp_size_t m,
	    mp_ptr tp, mp_size_t talloc)
{
  mp_size_t q0size;
  mp_size_t q1size;

  unsigned i;
  
  ASSERT(hgcd->asize > m);
  
  for (i = 0; hgcd->bsize > m; i++)
    {
      if (!lehmer_step(hgcd, quotients))
	if (!euclid_step(hgcd, quotients, m, tp, talloc))
	  break;
    }
  
  trace("lehmer_loop: %d iterations needed.\n", i);
  
  /* Now we should have A and B of approximately the right size. Two cases:
     
  1. hgcd->bsize <= m, and then we need to back up two steps (or more,
     if by chance also hgcd->asize <= m)
      
  2. hgcd->bsize > m, hgcd->r0size <= m (from the euclidian step
     check). We need to back up only one step.
  */
  
  if (hgcd->bsize > m)
    {
      /* FIXME: Q0 and R0 was already computed by euclid_step, but due
	 to the interface of this function, we need to recompute it
	 here. */

      q1size = mpn_jhgcd_tdiv(quotients->top, hgcd->r0p, &hgcd->r0size,
			      hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
      
      ASSERT(hgcd->r0size <= m);
      q0size = jhgcd_backup(hgcd, quotients, tp, talloc);
    }
  else
    {
      ASSERT(hgcd->bsize <= m);

      q1size = jhgcd_backup(hgcd, quotients, tp, talloc);
      q0size = jhgcd_backup(hgcd, quotients, tp, talloc);

      ASSERT(hgcd->r1size <= m);

      /* FIXME: The final adjustment of the sizes seems suboptimal. We
	 should be able to avoid some extra steps by carefully
	 selecting the number of bits that are given to hgcd2. */
      
      while (hgcd->r0size <= m)
	{
	  q1size = q0size;
	  q0size = jhgcd_backup(hgcd, quotients, tp, talloc);
	}
    }

  /* Now a,b,r[0] >= W^m, r[1] < W^m */
  ASSERT(hgcd->asize > m);
  ASSERT(hgcd->bsize > m);
  ASSERT(hgcd->r0size > m);
  ASSERT(hgcd->r1size <= m);

  if (jhgcd_check_q(hgcd, quotients->top, q0size, tp, talloc))
    {
      /* We can use r0 */
      trace("lehmer_loop: r0 can be used.\n");

      mpn_quotient_stack_push(quotients, q0size);
      
      if (jhgcd_check_q(hgcd, quotients->top, q1size, tp, talloc))
	{
	  /* We can use r1 */
	  trace("lehmer_loop: r1 can be used.\n");
	  mpn_quotient_stack_push(quotients, q1size);
	}
    }
}

mp_size_t
mpn_jhgcd_lehmer_itch(struct mpn_jhgcd *hgcd)
{
  /* Two quotients (asize), plus space needed for jhgcd_backup (alloc
     + qsize) or hgcd_update (alloc + 1). */
  return 3 * (hgcd->asize + 1)
    + hgcd->alloc + 1;
}

int
mpn_jhgcd_lehmer(struct mpn_jhgcd *hgcd,
		 struct mpn_quotient_stack *quotients,
		 mp_ptr tp, mp_size_t talloc)
{
  mp_size_t n = hgcd->asize;
  mp_size_t m = CHOOSE_M(n);
  ASSERT(m);

  if (hgcd->bsize <= m)
    return 0;

  /* Perform one division first, to ensure that we can return success,
     and won't try to underflow the stack in the backup phase */

  if (!euclid_step(hgcd, quotients, m, tp, talloc))
    return 0;

  /* Reduce A and B until B gets reasonably small. */  
  lehmer_loop(hgcd, quotients, m, tp, talloc);

  return 1;
}

static void *
xalloc(size_t size)
{
  void *p = malloc(size);
  if (!p)
    {
      trace("Virtual memory exhausted.\n");
      abort();
    }
  return p;
}

mp_size_t
mpn_fgcd(mp_ptr rp,
	 mp_srcptr ap, mp_size_t asize,
	 mp_srcptr bp, mp_size_t bsize)
{
  struct mpn_jhgcd hgcd;
  struct mpn_quotient_stack quotients;
  mp_size_t rsize;
  
  mp_size_t talloc;
  mp_size_t salloc;

  /* The spaced may well be too large for the stack. */
  mp_ptr tp;
  mp_ptr hlimbs;
  mp_ptr qlimbs;
  mp_size_t *qsizes;

  /* Check for small case first */
  if (bsize < FGCD_LEHMER_THRESHOLD
      /* Second condition is to ensure that we have enough space on the
	 stack */
      && asize < 300)
    {
      mp_ptr ac;
      mp_ptr bc;
      
      /* FIXME: Shouldn't use alloca directly */
      tp = alloca( (2 * asize + bsize + 3) * sizeof(mp_limb_t));
      ac = tp + asize + 1;
      MPN_COPY(ac, ap, asize);
      bc = ac + asize + 1;
      MPN_COPY(bc, bp, bsize);
      
      return mpn_gcd_jebelean(rp, ac, asize, bc, bsize, tp);
    }
  
  talloc = mpn_jhgcd_itch(asize, bsize);
  hlimbs = xalloc(talloc * sizeof(mp_limb_t));
  
  mpn_jhgcd_init(&hgcd, ap, asize, bp, bsize,
		 hlimbs, talloc);
  
  talloc = mpn_quotient_stack_limb_need(asize, bsize);
  salloc = mpn_quotient_stack_size_need(asize, bsize);

  qlimbs = xalloc(talloc * sizeof(mp_limb_t));
  qsizes = xalloc(salloc * sizeof(mp_size_t));
  
  mpn_quotient_stack_init(&quotients,
			  qlimbs, talloc, qsizes, salloc);
  
  talloc = mpn_jhgcd_schoenhage_itch(&hgcd);
  tp = xalloc(talloc * sizeof(mp_limb_t));
  
  for (;;)
    {
      trace("mpn_fgcd: a = %Nd\n"
	    "          b = %Nd\n",
	    hgcd.ap, hgcd.asize, hgcd.bp, hgcd.bsize);

      ASSERT(MPN_LESS_P(hgcd.bp, hgcd.bsize, hgcd.ap, hgcd.asize));
	     
      if (hgcd.bsize < FGCD_LEHMER_THRESHOLD)
	{
	  trace("mpn_fgcd: x is small, falling back to jebelean.\n");	  
	  rsize = mpn_gcd_jebelean(rp, hgcd.ap, hgcd.asize,
				   hgcd.bp, hgcd.bsize, tp);

	  break;
	}

      /* Reset matrix. */
      mpn_jhgcd_identity(&hgcd);
      
      if (!mpn_jhgcd_schoenhage(&hgcd,
				&quotients,
				tp, talloc))
	{
	  trace("mpn_fgcd: mpn_jhgcd_schoenhage returned zero.\n");
	  if (!euclid_step(&hgcd, &quotients, 0, tp, talloc))
	    {
	      /* a % b == 0 */
	      trace("mpn_fgcd: euclid step gave a zero remainder.\n");
	      MPN_COPY(rp, hgcd.bp, hgcd.bsize);

	      rsize = hgcd.bsize;
	      break;
	    }    
	}
    }
  free(tp);
  free(hlimbs);
  free(qlimbs);
  free(qsizes);

  return rsize;
}

/* Computes R = 2^k H + u A' - v B', which must be non-negative. */

static mp_size_t
jhgcd_fix(mp_size_t k,
	 mp_ptr rp, mp_size_t ralloc,
	 mp_ptr hp, mp_size_t hsize,
	 mp_srcptr up, mp_size_t usize, mp_srcptr ap,
	 mp_srcptr vp, mp_size_t vsize, mp_srcptr bp,
	 mp_ptr tp, mp_size_t talloc)
{
  mp_size_t tsize;
  mp_limb_t cy;
  mp_size_t asize = k;
  mp_size_t bsize = k;
  mp_size_t rsize;

  MPN_NORMALIZE(ap, asize);
  MPN_NORMALIZE(bp, bsize);

  tsize = MPN_MUL_SIZE(asize, usize);

  ASSERT(k + hsize <= ralloc);
  ASSERT(tsize <= talloc);
  ASSERT(tsize <= ralloc);
  /* FIXME: Figure out exactly what sizes we can get */
  ASSERT(usize <= k + 1);
  
  if (tsize)
    {
      tsize -= mpn_mul_friendly(tp,
				ap, asize,
				up, usize);
      
      trace("jhgcd_fix: k = %d, tsize = %d, talloc = %d,\n"
	    "  usize = %d, vsize = %d\n",
	    k, tsize, talloc, usize, vsize);

      if (tsize <= k)
	{
	  MPN_COPY(rp, tp, tsize);
	  MPN_COPY(rp + k, hp, hsize);
	  rsize = k + hsize;

	  if (tsize < k)
	    MPN_ZERO(rp + tsize, k - tsize);
	}
      else
	{
	  MPN_COPY(rp, tp, k);
	  rsize = k + mpn_add_friendly(rp + k,
				       hp, hsize,
				       tp + k, tsize - k);
	}
    }
  else
    {
      MPN_ZERO(rp, k);
      MPN_COPY(rp + k, hp, hsize);
      rsize = k + hsize;
    }

  ASSERT(rsize <= ralloc);
  
  /* A -= v * B' */
  tsize = MPN_MUL_SIZE(bsize, vsize);

  ASSERT(tsize <= talloc);
  ASSERT(tsize <= ralloc);
  /* FIXME: Figure out exactly what sizes we can get */
  ASSERT(vsize <= k + 1);

  if (tsize)
    {
      tsize -= mpn_mul_friendly(tp,
				bp, bsize,
				vp, vsize);

      ASSERT(tsize <= rsize);
      
      cy = mpn_sub(rp, rp, rsize,
		   tp, tsize);
      ASSERT(!cy);

      MPN_NORMALIZE(rp, rsize);
    }
  return rsize;
}

static int
schoenhage_step(struct mpn_jhgcd *hgcd,
		struct mpn_quotient_stack *quotients,
		mp_size_t k,
		mp_ptr tp, mp_size_t talloc)
{  
  struct mpn_jhgcd S;
  mp_size_t tsize;

  ASSERT(k < hgcd->asize);

  if (hgcd->bsize <= k)
    return 0;

  if (hgcd->asize < k + HGCD_LIMIT)
    return 0;

  trace("schoenhage_step: before jhgcd: asize = %d, bsize = %d, k = %d\n"
	"  a = %Nd, b = %Nd\n",
	hgcd->asize, hgcd->bsize, k,
	hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);

  tsize = mpn_jhgcd_itch(hgcd->asize - k, hgcd->bsize - k);
  ASSERT(tsize < talloc);

  mpn_jhgcd_init(&S,
		 hgcd->ap + k, hgcd->asize - k,
		 hgcd->bp + k, hgcd->bsize - k,
		 tp, tsize);

  tp += tsize; talloc -= tsize;
  
  if (talloc < mpn_jhgcd_schoenhage_itch(&S))
    trace("schoenhage_step: We're getting short on scratch space.\n");

  if (!mpn_jhgcd_schoenhage(&S, quotients, tp, talloc))
    /* B' already small */
    return 0;

#if 0
  trace("schoenhage_step: after jhgcd: asize = %d, bsize = %d\n"
	"  a = %Nd, b = %Nd\n"
	"  uvsize = (%d, %d; %d, %d)\n"
	"  uv = %s(%Nd, %Nd; %Nd, %Nd)\n",
	S.asize, S.bsize,
	S.ap, S.asize, S.bp, S.bsize,
	S.uvsize[0], S.uvsize[1], S.uvsize[2], S.uvsize[3],
	S.sign < 0 ? "-" : "",
	S.uvp[0], S.uvsize[0], S.uvp[1], S.uvsize[1],
	S.uvp[2], S.uvsize[2], S.uvp[3], S.uvsize[3]);
#endif
  
  /* Store new values in r0 and r1 */
  if (S.sign >= 0)
    {
      hgcd->r0size = jhgcd_fix(k, hgcd->r0p, hgcd->bsize,
			       S.ap, S.asize,
			       S.uvp[0], S.uvsize[0], hgcd->ap, 
			       S.uvp[1], S.uvsize[1], hgcd->bp,
			       tp, talloc);
      
      hgcd->r1size = jhgcd_fix(k, hgcd->r1p, hgcd->bsize,
			       S.bp, S.bsize,
			       S.uvp[3], S.uvsize[3], hgcd->bp,
			       S.uvp[2], S.uvsize[2], hgcd->ap,
			       tp, talloc); 
    }
  else
    {
      hgcd->r0size = jhgcd_fix(k, hgcd->r0p, hgcd->asize,
			       S.ap, S.asize,
			       S.uvp[1], S.uvsize[1], hgcd->bp,
			       S.uvp[0], S.uvsize[0], hgcd->ap,
			       tp, talloc);
      
      hgcd->r1size = jhgcd_fix(k, hgcd->r1p, hgcd->bsize,
			       S.bp, S.bsize,
			       S.uvp[2], S.uvsize[2], hgcd->ap,
			       S.uvp[3], S.uvsize[3], hgcd->bp,
			       tp, talloc);
    }

  MP_PTR_SWAP(hgcd->ap, hgcd->r0p);
  MP_PTR_SWAP(hgcd->bp, hgcd->r1p);
  hgcd->asize = hgcd->r0size;
  hgcd->bsize = hgcd->r1size;
  
  trace("schoenhage_step, reduced: a = %Nd, b = %Nd\n",
	hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  ASSERT(MPN_LESS_P(hgcd->bp, hgcd->bsize, hgcd->ap, hgcd->asize));

  mpn_jhgcd_mul(&S, hgcd, tp, talloc);
  return 1;
}

mp_size_t
mpn_jhgcd_schoenhage_itch(struct mpn_jhgcd *hgcd)
{
  return 2*mpn_jhgcd_lehmer_itch(hgcd)
    + mpn_jhgcd_itch(hgcd->asize, hgcd->bsize);
}

#ifdef mpn_jhgcd_schoenhage
# undef mpn_jhgcd_schoenhage
#endif

int
mpn_jhgcd_schoenhage(struct mpn_jhgcd *hgcd,
		     struct mpn_quotient_stack *quotients,
		     mp_ptr tp, mp_size_t talloc)
{
  mp_size_t n = hgcd->asize;
  mp_size_t m = CHOOSE_M(n);

  ASSERT(m);

  if (n < HGCD_LEHMER_THRESHOLD)
    return mpn_jhgcd_lehmer(hgcd, quotients, tp, talloc);
  
  if (hgcd->bsize <= m)
    return 0;

  /* Perform one division first, to ensure that we can return success,
     and won't try to underflow the stack in the backup phase */

  if (!euclid_step(hgcd, quotients, m, tp, talloc))
    return 0;
    
  if ( (schoenhage_step(hgcd, quotients, n / 2, tp, talloc)
	|| euclid_step(hgcd, quotients, m, tp, talloc))
       && hgcd->bsize > m)
    {
      trace("mpn_jhgcd_schoenhage: after first schoenhage step:\n"
	    "  n = %d, asize = %d, bsize = %d\n",
	    n, hgcd->asize, hgcd->bsize);
      
      /* FIXME: Always perform one euclid step here? */

      ASSERT(hgcd->asize < 2*m + 1);
      schoenhage_step(hgcd, quotients, 2 * m + 1 - hgcd->asize, tp, talloc);
    }

  /* FIXME: The final adjustment of the sizes seems suboptimal. We
   * should be able to avoid some extra steps by carefully selecting
   * the number of bits that are given to hgcd2. */
  
  lehmer_loop(hgcd, quotients, m, tp, talloc);
  
  return 1;
}

#if WANT_CHECKS
int
mpn_jhgcd_schoenhage_wrapper(struct mpn_jhgcd *hgcd,
			     struct mpn_quotient_stack *quotients,
			     mp_ptr tp, mp_size_t talloc)
{
  struct mpn_jhgcd check_hgcd;
  struct mpn_quotient_stack check_quotients;
  mp_size_t check_talloc;
  mp_size_t check_salloc;
  int check_result;
  int result;

  static int count = 0;
  int entry = ++count;

  trace("mpn_jhgcd_schoenhage_wrapper: %d\n"
	"  a = %Nd, b = %Nd\n",
	entry, hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize);
  
  check_talloc = mpn_jhgcd_itch(hgcd->asize, hgcd->bsize);
  mpn_jhgcd_init(&check_hgcd,
		 hgcd->ap, hgcd->asize, hgcd->bp, hgcd->bsize,
		 alloca(check_talloc * sizeof(mp_limb_t)), check_talloc);

  check_talloc = mpn_quotient_stack_limb_need(hgcd->asize, hgcd->bsize);
  check_salloc = mpn_quotient_stack_size_need(hgcd->asize, hgcd->bsize);

  mpn_quotient_stack_init(&check_quotients,
			  alloca(check_talloc * sizeof(mp_limb_t)),
			  check_talloc,
			  alloca(check_salloc * sizeof(mp_size_t)),
			  check_salloc);
  
  check_talloc = mpn_jhgcd_lehmer_itch(&check_hgcd);
  check_result = mpn_jhgcd_lehmer(&check_hgcd, &check_quotients, 
				  alloca(check_talloc * sizeof(mp_limb_t)),
				  check_talloc);

  result = mpn_jhgcd_schoenhage(hgcd, quotients, tp, talloc);

  if (result != check_result)
    {
      trace("mpn_jhgcd_schoenhage_wrapper: %d\n"
	    "  mpn_jhgcd_schoenhage returned %d, expected %d\n",
	    entry, result, check_result);
      abort();
    }

  if (!mpn_jhgcd_equal(hgcd, &check_hgcd))
    {
      trace("mpn_jhgcd_schoenhage_wrapper: %d\n"
	    "  mpn_jhgcd_schoenhage didn't return the right array.\n",
	    entry);
      trace_mpn_matrix("got", hgcd);
      trace_mpn_matrix("expected", &check_hgcd);

      trace_mpn_quotients("got quotients", quotients);
      trace_mpn_quotients("expected", &check_quotients);
      
      abort();
    }

  trace("mpn_jhgcd_schoenhage_wrapper: %d\n"
	"  returning %d\n",
	entry, result);
  
  /* FIXME: Check quotients as well. But that's a little more complicated */
  return result;
}
#endif
